@@ -12,13 +12,9 @@
 #include "axfr.h"
 #include "dns.h"
 #include "packet.h"
+#include "options.h"
 
-#ifdef LIBWRAP
-#include <tcpd.h>
-
-int allow_severity = LOG_INFO;
-int deny_severity = LOG_NOTICE;
-#endif /* LIBWRAP */
+#define AXFR_TSIG_SIGN_EVERY_NTH	96	/* tsig sign every N packets. */
 
 query_state_type
 query_axfr (struct nsd *nsd, struct query *query)
@@ -36,6 +32,16 @@ query_axfr (struct nsd *nsd, struct query *query)
 		query->maxlen = AXFR_MAX_MESSAGE_LEN;
 	
 	assert(!query_overflow(query));
+#ifdef TSIG
+	/* only keep running values for most packets */
+	query->tsig_prepare_it = 0;
+	query->tsig_update_it = 1;
+	if(query->tsig_sign_it) {
+		/* prepare for next updates */
+		query->tsig_prepare_it = 1;
+		query->tsig_sign_it = 0;
+	}
+#endif /* TSIG */
 
 	if (query->axfr_zone == NULL) {
 		/* Start AXFR.  */
@@ -57,9 +63,14 @@ query_axfr (struct nsd *nsd, struct query *query)
 		}
 
 		query->axfr_current_domain
-			= (domain_type *) heap_first(nsd->db->domains->names_to_domains);
+			= (domain_type *) rbtree_first(nsd->db->domains->names_to_domains);
 		query->axfr_current_rrset = NULL;
 		query->axfr_current_rr = 0;
+#ifdef TSIG
+		if(query->tsig.status == TSIG_OK) {
+			query->tsig_sign_it = 1; /* sign first packet in stream */
+		}
+#endif /* TSIG */
 
 		query_add_compression_domain(query, query->domain, QHEADERSZ);
 
@@ -86,7 +97,7 @@ query_axfr (struct nsd *nsd, struct query *query)
 	/* Add zone RRs until answer is full.  */
 	assert(query->axfr_current_domain);
 	
-	while ((rbnode_t *) query->axfr_current_domain != HEAP_NULL) {
+	while ((rbnode_t *) query->axfr_current_domain != RBTREE_NULL) {
 		if (!query->axfr_current_rrset) {
 			query->axfr_current_rrset = domain_find_any_rrset(
 				query->axfr_current_domain,
@@ -114,7 +125,7 @@ query_axfr (struct nsd *nsd, struct query *query)
 		}
 		assert(query->axfr_current_domain);
 		query->axfr_current_domain
-			= (domain_type *) heap_next((rbnode_t *) query->axfr_current_domain);
+			= (domain_type *) rbtree_next((rbnode_t *) query->axfr_current_domain);
 	}
 
 	/* Add terminating SOA RR.  */
@@ -124,6 +135,9 @@ query_axfr (struct nsd *nsd, struct query *query)
 				 &query->axfr_zone->soa_rrset->rrs[0]);
 	if (added) {
 		++total_added;
+#ifdef TSIG
+		query->tsig_sign_it = 1; /* sign last packet */
+#endif /* TSIG */
 		query->axfr_is_done = 1;
 	}
 
@@ -131,6 +145,15 @@ return_answer:
 	ANCOUNT_SET(query->packet, total_added);
 	NSCOUNT_SET(query->packet, 0);
 	ARCOUNT_SET(query->packet, 0);
+
+#ifdef TSIG
+	/* check if it needs tsig signatures */
+	if(query->tsig.status == TSIG_OK) {
+		if(query->tsig.updates_since_last_prepare >= AXFR_TSIG_SIGN_EVERY_NTH) {
+			query->tsig_sign_it = 1;
+		}
+	}
+#endif /* TSIG */
 	query_clear_compression_tables(query);
 	return QUERY_IN_AXFR;
 }
@@ -141,52 +164,27 @@ return_answer:
 query_state_type
 answer_axfr_ixfr(struct nsd *nsd, struct query *q)
 {
+	acl_options_t *acl;
 	/* Is it AXFR? */
 	switch (q->qtype) {
 	case TYPE_AXFR:
-#ifndef DISABLE_AXFR		/* XXX Should be a run-time flag */
 		if (q->tcp) {
-#ifdef LIBWRAP
-			struct request_info request;
-#ifdef AXFR_DAEMON_PREFIX
-			const uint8_t *qptr = dname_name(q->qname);
-			char axfr_daemon[MAXDOMAINLEN + sizeof(AXFR_DAEMON_PREFIX)];
-			char *t = axfr_daemon + sizeof(AXFR_DAEMON_PREFIX) - 1;
-
-			memcpy(axfr_daemon, AXFR_DAEMON_PREFIX, sizeof(AXFR_DAEMON_PREFIX));
-
-			/* Copy the qname as a string */
-			while (*qptr)
+			zone_options_t* zone_opt;
+			zone_opt = zone_options_find(nsd->options, q->qname);
+			if(!zone_opt ||
+			   acl_check_incoming(zone_opt->provide_xfr, q, &acl)==-1) 
 			{
-				memcpy(t, qptr + 1, *qptr);
-				t += *qptr;
-				*t++ = '.';
-				qptr += *qptr + 1;
+				DEBUG(DEBUG_XFRD,1, (LOG_INFO, "axfr refused, %s", 
+						acl?"blocked":"no acl matches"));
+				RCODE_SET(q->packet, RCODE_REFUSE);
+				return QUERY_PROCESSED;
 			}
-			*t = 0;
-			
-#endif /* AXFR_DAEMON_PREFIX */
-			request_init(&request, RQ_DAEMON, AXFR_DAEMON, RQ_CLIENT_SIN, &q->addr, 0);
-			sock_methods(&request);	/* This is to work around the bug in libwrap */
-			if (!hosts_access(&request)) {
-#ifdef AXFR_DAEMON_PREFIX
-				request_init(&request, RQ_DAEMON, axfr_daemon, RQ_CLIENT_SIN, &q->addr, 0);
-				sock_methods(&request);	/* This is to work around the bug in libwrap */
-				log_msg(LOG_ERR, "checking %s", axfr_daemon);
-				if (!hosts_access(&request)) {
-#endif /* AXFR_DAEMON_PREFIX */
-					RCODE_SET(q->packet, RCODE_REFUSE);
-					return QUERY_PROCESSED;
-#ifdef AXFR_DAEMON_PREFIX
-				}
-#endif /* AXFR_DAEMON_PREFIX */
-			}
-#endif /* LIBWRAP */
+			DEBUG(DEBUG_XFRD,1, (LOG_INFO, "axfr admitted acl %s %s", 
+				acl->ip_address_spec, acl->key_name?acl->key_name:"NOKEY"));
 			return query_axfr(nsd, q);
 		}
-#endif	/* DISABLE_AXFR */
 	case TYPE_IXFR:
-		RCODE_SET(q->packet, RCODE_REFUSE);
+		RCODE_SET(q->packet, RCODE_IMPL);
 		return QUERY_PROCESSED;
 	default:
 		return QUERY_DISCARDED;
