@@ -66,7 +66,8 @@
      VAR_KEY = 292,
      VAR_ALGORITHM = 293,
      VAR_SECRET = 294,
-     VAR_AXFR = 295
+     VAR_AXFR = 295,
+     VAR_VERBOSITY = 296
    };
 #endif
 /* Tokens.  */
@@ -108,6 +109,7 @@
 #define VAR_ALGORITHM 293
 #define VAR_SECRET 294
 #define VAR_AXFR 295
+#define VAR_VERBOSITY 296
 
 
 
@@ -118,7 +120,7 @@ typedef union YYSTYPE {
 	char*	str;
 } YYSTYPE;
 /* Line 1447 of yacc.c.  */
-#line 122 "configparser.h"
+#line 124 "configparser.h"
 # define yystype YYSTYPE /* obsolescent; will be withdrawn */
 # define YYSTYPE_IS_DECLARED 1
 # define YYSTYPE_IS_TRIVIAL 1
