@@ -31,12 +31,9 @@ detect_nsec3_params(rr_type* nsec3_apex,
 	/* always uses first NSEC3 record with SOA bit set */
 	assert(salt && salt_len && iter);
 	assert(nsec3_apex);
-	*salt_len = rdata_atom_data(nsec3_apex->rdatas[2])[0];
-	*salt = (unsigned char*)(rdata_atom_data(nsec3_apex->rdatas[2])+1);
-	*iter = (rdata_atom_data(nsec3_apex->rdatas[1])[0]<<16) |
-		(rdata_atom_data(nsec3_apex->rdatas[1])[1]<<8) |
-		(rdata_atom_data(nsec3_apex->rdatas[1])[2]);
-	*iter &= 0x7fffff;
+	*salt_len = rdata_atom_data(nsec3_apex->rdatas[3])[0];
+	*salt = (unsigned char*)(rdata_atom_data(nsec3_apex->rdatas[3])+1);
+	*iter = read_uint16(rdata_atom_data(nsec3_apex->rdatas[2]));
 }
 
 static const dname_type *
@@ -70,10 +67,10 @@ nsec3_hash_dname(region_type *region, zone_type *zone,
 static int 
 nsec3_has_soa(rr_type* rr)
 {
-	if(rdata_atom_size(rr->rdatas[4]) > 0 && /* has types in bitmap */
-		rdata_atom_data(rr->rdatas[4])[0] == 0 && /* first window = 0, */
+	if(rdata_atom_size(rr->rdatas[5]) > 0 && /* has types in bitmap */
+		rdata_atom_data(rr->rdatas[5])[0] == 0 && /* first window = 0, */
 						/* [1]: windowlen must be >= 1 */
-		rdata_atom_data(rr->rdatas[4])[2]&0x02)  /* SOA bit set */
+		rdata_atom_data(rr->rdatas[5])[2]&0x02)  /* SOA bit set */
 		return 1;
 	return 0;
 }
@@ -104,9 +101,14 @@ find_zone_nsec3(namedb_type* namedb, zone_type *zone)
 				rdata_atom_data(rr->rdatas[0])[0]);
 			continue;
 		}
+		if(rdata_atom_data(rr->rdatas[1])[0] != 0) {
+			/* draft-nsec3-09: NSEC3PARAM records with flags
+			   field value other than zero MUST be ignored. */
+			continue;
+		}
 		/* check hash of apex -> NSEC3 with soa bit on */
 		hashed_apex = nsec3_hash_dname_param(tmpregion,
-			zone, domain_dname(zone->apex), paramset->rrs);
+			zone, domain_dname(zone->apex), &paramset->rrs[i]);
 		domain = domain_table_find(namedb->domains, hashed_apex);
 		if(!domain) {
 			log_msg(LOG_ERR, "%s NSEC3PARAM entry %d has no hash(apex).", 
@@ -137,10 +139,10 @@ find_zone_nsec3(namedb_type* namedb, zone_type *zone)
 					rdata_atom_data(nsec3_rrset->rrs[j].rdatas[0])[0]
 				&& memcmp(salt1, salt2, saltlen1) == 0) {
 				/* found it */
-				log_msg(LOG_INFO, 
+				DEBUG(DEBUG_QUERY, 1, (LOG_INFO, 
 					"detected NSEC3 for zone %s saltlen=%d iter=%d",
-						dname_to_string(domain_dname(
-						zone->apex),0), saltlen2, iter2);
+					dname_to_string(domain_dname(
+					zone->apex),0), saltlen2, iter2));
 				region_destroy(tmpregion);
 				return &nsec3_rrset->rrs[j];
 			}
@@ -213,16 +215,14 @@ nsec3_rrset_params_ok(rr_type* base, rrset_type* rrset)
 		assert(rrset->rrs[i].type == TYPE_NSEC3);
 		if(rdata_atom_data(rd[0])[0] == 
 			rdata_atom_data(prd[0])[0] && /* hash algo */
-		   (rdata_atom_data(rd[1])[0]&0x7f) == 
-			(rdata_atom_data(prd[1])[0]&0x7f) && /* iterations 0 */
-		   rdata_atom_data(rd[1])[1] == 
-			rdata_atom_data(prd[1])[1] && /* iterations 1 */
-		   rdata_atom_data(rd[1])[2] == 
-			rdata_atom_data(prd[1])[2] && /* iterations 2 */
 		   rdata_atom_data(rd[2])[0] == 
-			rdata_atom_data(prd[2])[0] && /* salt length */
-		   memcmp(rdata_atom_data(rd[2])+1, 
-			rdata_atom_data(prd[2])+1, rdata_atom_data(rd[2])[0]) 
+			rdata_atom_data(prd[2])[0] && /* iterations 0 */
+		   rdata_atom_data(rd[2])[1] == 
+			rdata_atom_data(prd[2])[1] && /* iterations 1 */
+		   rdata_atom_data(rd[3])[0] == 
+			rdata_atom_data(prd[3])[0] && /* salt length */
+		   memcmp(rdata_atom_data(rd[3])+1, 
+			rdata_atom_data(prd[3])+1, rdata_atom_data(rd[3])[0]) 
 			== 0 ) 
 		{
 			/* this NSEC3 matches nsec3 parameters from zone */
@@ -232,26 +232,6 @@ nsec3_rrset_params_ok(rr_type* base, rrset_type* rrset)
 	return 0;
 }
 
-static domain_type* 
-nsec3_find_last(zone_type* zone)
-{
-	/* this is the longest possible walk to get to the end, but is precomputed */
-	/* could also tree-find the <max_val>.domain name */
-	domain_type* walk = zone->apex;
-	domain_type* result = 0;
-	while(walk && 
-		dname_is_subdomain(domain_dname(walk), domain_dname(zone->apex)))
-	{
-		/* remember last domain with an OK NSEC3 rrset */
-		if(nsec3_rrset_params_ok(NULL,
-			domain_find_rrset(walk, zone, TYPE_NSEC3))) {
-			result = walk;
-		}
-		walk = domain_next(walk);
-	}
-	return result;
-}
-
 int 
 nsec3_find_cover(namedb_type* db, zone_type* zone, 
 	const dname_type* hashname, domain_type** result)
@@ -277,7 +257,8 @@ nsec3_find_cover(namedb_type* db, zone_type* zone,
 	}
 
 	/* find covering NSEC3 record, lexicographically before the closest match */
-	walk = closest_match;
+	/* use nsec3_lookup to jumpstart the search */
+	walk = closest_match->nsec3_lookup;
 	rrset = 0;
 	while(walk && dname_is_subdomain(domain_dname(walk), domain_dname(zone->apex)))
 	{
@@ -318,9 +299,9 @@ prehash_domain(namedb_type* db, zone_type* zone,
 	if(!zone->nsec3_soa_rr)
 	{
 		/* set to 0 (in case NSEC3 removed after an update) */
-		domain->nsec3_exact = 0;
-		domain->nsec3_cover = 0;
-		domain->nsec3_wcard_child_cover = 0;
+		domain->nsec3_is_exact = 0;
+		domain->nsec3_cover = NULL;
+		domain->nsec3_wcard_child_cover = NULL;
 		return;
 	}
 
@@ -328,8 +309,8 @@ prehash_domain(namedb_type* db, zone_type* zone,
 	exact = nsec3_find_cover(db, zone, hashname, &result);
 	domain->nsec3_cover = result;
 	if(exact)
-		domain->nsec3_exact = result;
-	else	domain->nsec3_exact = 0;
+		domain->nsec3_is_exact = 1;
+	else	domain->nsec3_is_exact = 0;
 
 	if(!exact && domain->is_existing && !domain_has_only_NSEC3(domain, zone)) {
 		/* make sure it is not glue */
@@ -386,7 +367,7 @@ prehash_ds(namedb_type* db, zone_type* zone,
 	int exact;
 
 	if(!zone->nsec3_soa_rr) {
-		domain->nsec3_ds_parent_exact = NULL;
+		domain->nsec3_ds_parent_is_exact = 0;
 		domain->nsec3_ds_parent_cover = NULL;
 		return;
 	}
@@ -395,8 +376,8 @@ prehash_ds(namedb_type* db, zone_type* zone,
 	hashname = nsec3_hash_dname(region, zone, domain_dname(domain));
 	exact = nsec3_find_cover(db, zone, hashname, &result);
 	if(exact)
-		domain->nsec3_ds_parent_exact = result;
-	else 	domain->nsec3_ds_parent_exact = 0;
+		domain->nsec3_ds_parent_is_exact = 1;
+	else 	domain->nsec3_ds_parent_is_exact = 0;
 	domain->nsec3_ds_parent_cover = result;
 
 	/*
@@ -409,6 +390,7 @@ static void
 prehash_zone(struct namedb* db, struct zone* zone)
 {
 	domain_type *walk;
+	domain_type *last_nsec3_node;
 	region_type *temp_region = region_create(xalloc, free);
 	assert(db && zone);
 
@@ -416,9 +398,6 @@ prehash_zone(struct namedb* db, struct zone* zone)
 	zone->nsec3_soa_rr = find_zone_nsec3(db, zone);
 	if(!zone->nsec3_soa_rr) 
 		zone->nsec3_last = 0;
-	else	zone->nsec3_last = nsec3_find_last(zone); 
-	assert((zone->nsec3_soa_rr&&zone->nsec3_last) ||
-		(!zone->nsec3_soa_rr&&!zone->nsec3_last));
 	if(zone->nsec3_soa_rr) {
 		/* check that hashed, the apex name equals the found nsec3 domain */
 		const dname_type* checkname = nsec3_hash_dname(temp_region, 
@@ -434,6 +413,27 @@ prehash_zone(struct namedb* db, struct zone* zone)
 			zone->nsec3_last = 0;
 		}
 	}
+	if(!zone->nsec3_soa_rr)
+		return;
+
+	/* go through entire zone and setup nsec3_lookup speedup */
+	walk = zone->apex;
+	last_nsec3_node = NULL;
+	/* since we walk in sorted order, we pass all NSEC3s in sorted
+	   order and we can set the lookup ptrs */
+	while(walk && dname_is_subdomain(
+		domain_dname(walk), domain_dname(zone->apex)))
+	{
+		zone_type* z = domain_find_zone(walk);
+		if(z && z==zone)
+		{
+			if(domain_find_rrset(walk, zone, TYPE_NSEC3))
+				last_nsec3_node = walk;
+			walk->nsec3_lookup = last_nsec3_node;
+		}
+		walk = domain_next(walk);
+	}
+	zone->nsec3_last = last_nsec3_node;
 
 	/* go through entire zone */
 	walk = zone->apex;
@@ -462,20 +462,23 @@ prehash_zone(struct namedb* db, struct zone* zone)
 }
 
 void 
-prehash(struct namedb* db, struct zone* zone)
+prehash(struct namedb* db, int updated_only)
 {
-	time_t start = time(0), end;
-	if(zone) {
-		prehash_zone(db, zone);
-	} else {
-		zone_type *z;
-		for(z = db->zones; z; z = z->next)
-		{
+	zone_type *z;
+	time_t end, start = time(NULL);
+	int count = 0;
+	for(z = db->zones; z; z = z->next)
+	{
+		if(!updated_only || z->updated) {
 			prehash_zone(db, z);
+			if(z->nsec3_soa_rr)
+				count++;
 		}
 	}
-	end = time(0);
-	log_msg(LOG_INFO, "prehash took %d seconds", (int)(end-start));
+	end = time(NULL);
+	if(count > 0)
+		DEBUG(DEBUG_QUERY, 1, (LOG_INFO, "nsec3-prepare took %d "
+		"seconds for %d zones.", (int)(end-start), count));
 }
 
 /* add the NSEC3 rrset to the query answer at the given domain */
@@ -531,7 +534,9 @@ nsec3_add_closest_encloser_proof(
 	/* prove that below closest encloser nothing exists */
 	nsec3_add_nonexist_proof(query, answer, closest_encloser, db, qname);
 	/* proof that closest encloser exists */
-	nsec3_add_rrset(query, answer, AUTHORITY_SECTION, closest_encloser->nsec3_exact);
+	if(closest_encloser->nsec3_is_exact)
+		nsec3_add_rrset(query, answer, AUTHORITY_SECTION, 
+			closest_encloser->nsec3_cover);
 }
 
 void 
@@ -554,27 +559,27 @@ nsec3_add_ds_proof(struct query *query, struct answer *answer,
 	/*
 	printf("Add ds proof for %s\n", dname_to_string(domain_dname(domain),0));
 	*/
-	if(domain->nsec3_ds_parent_exact) {
+	if(domain->nsec3_ds_parent_is_exact) {
 		/* use NSEC3 record from above the zone cut. */
 		nsec3_add_rrset(query, answer, AUTHORITY_SECTION, 
-			domain->nsec3_ds_parent_exact);
+			domain->nsec3_ds_parent_cover);
 	} else {
 		/* prove closest provable encloser */
 		domain_type* par = domain->parent;
 		domain_type* prev_par = 0;
-		while(par && !par->nsec3_exact)
+		while(par && !par->nsec3_is_exact)
 		{
 			prev_par = par;
 			par = par->parent;
 		}
 		assert(par); /* parent zone apex must be provable, thus this ends */	
 		nsec3_add_rrset(query, answer, AUTHORITY_SECTION,
-			par->nsec3_exact);
+			par->nsec3_cover);
 		/* we took several steps to go to the provable parent, so
 		   the one below it has no exact nsec3, disprove it.
 		   disprove is easy, it has a prehashed cover ptr. */
 		if(prev_par) {
-			assert(prev_par != domain && !prev_par->nsec3_exact);
+			assert(prev_par != domain && !prev_par->nsec3_is_exact);
 			nsec3_add_rrset(query, answer, AUTHORITY_SECTION,
 				prev_par->nsec3_cover);
 		}
@@ -597,8 +602,9 @@ nsec3_answer_nodata(struct query *query, struct answer *answer,
 		if(original == query->zone->apex) {
 			/* DS at zone apex, but server not authoritative for parent zone */
 			/* so answer at the child zone level */
-			nsec3_add_rrset(query, answer, AUTHORITY_SECTION, 
-				original->nsec3_exact);
+			if(original->nsec3_is_exact)
+				nsec3_add_rrset(query, answer, AUTHORITY_SECTION, 
+					original->nsec3_cover);
 			return;
 		}
 		/* query->zone must be the parent zone */
@@ -609,16 +615,18 @@ nsec3_answer_nodata(struct query *query, struct answer *answer,
 		&& label_is_wildcard(dname_name(domain_dname(original)))) {
 		/* denial for wildcard is already there */
 		/* add parent proof to have a closest encloser proof for wildcard parent */
-		if(original->parent)
+		if(original->parent && original->parent->nsec3_is_exact)
 			nsec3_add_rrset(query, answer, AUTHORITY_SECTION, 
-				original->parent->nsec3_exact);
+				original->parent->nsec3_cover);
 		/* proof for wildcard itself */
 		nsec3_add_rrset(query, answer, AUTHORITY_SECTION, 
 			original->nsec3_cover);
 	}
-	else	/* add nsec3 to prove rrset does not exist */
-		nsec3_add_rrset(query, answer, AUTHORITY_SECTION, 
-			original->nsec3_exact);
+	else {	/* add nsec3 to prove rrset does not exist */
+		if(original->nsec3_is_exact)
+			nsec3_add_rrset(query, answer, AUTHORITY_SECTION, 
+				original->nsec3_cover);
+	}
 }
 
 void 
@@ -637,7 +645,7 @@ domain_has_only_NSEC3(struct domain* domain, struct zone* zone)
 	int nsec3_seen = 0, rrsig_seen = 0;
 	while(rrset)
 	{
-		if(rrset->zone == zone)
+		if(!zone || rrset->zone == zone)
 		{
 			if(rrset->rrs[0].type == TYPE_NSEC3)
 				nsec3_seen = 1;
@@ -669,7 +677,9 @@ nsec3_answer_authoritative(struct domain** match, struct query *query,
 		/* act as if the NSEC3 domain did not exist, name error */
 		*match = 0;
 		/* all nsec3s are directly below the apex, that is closest encloser */
-		nsec3_add_rrset(query, answer, AUTHORITY_SECTION, query->zone->apex->nsec3_exact);
+		if(query->zone->apex->nsec3_is_exact)
+			nsec3_add_rrset(query, answer, AUTHORITY_SECTION, 
+				query->zone->apex->nsec3_cover);
 		/* disprove the nsec3 record. */
 		nsec3_add_rrset(query, answer, AUTHORITY_SECTION, closest_encloser->nsec3_cover);
 		/* disprove a wildcard */
