@@ -389,6 +389,12 @@ server_init(struct nsd *nsd)
 # endif
 		}
 #endif
+		/* set it nonblocking */
+		/* otherwise, on OSes with thundering herd problems, the
+		   UDP recv could block NSD after select returns readable. */
+		if (fcntl(nsd->udp[i].s, F_SETFL, O_NONBLOCK) == -1) {
+			log_msg(LOG_ERR, "cannot fcntl udp: %s", strerror(errno));
+		}
 
 		/* Bind it... */
 		if (bind(nsd->udp[i].s, (struct sockaddr *) nsd->udp[i].addr->ai_addr, nsd->udp[i].addr->ai_addrlen) != 0) {
@@ -432,6 +438,12 @@ server_init(struct nsd *nsd)
 			return -1;
 		}
 #endif
+		/* set it nonblocking */
+		/* (StevensUNP p463), if tcp listening socket is blocking, then
+		   it may block in accept, even if select() says readable. */
+		if (fcntl(nsd->tcp[i].s, F_SETFL, O_NONBLOCK) == -1) {
+			log_msg(LOG_ERR, "cannot fcntl tcp: %s", strerror(errno));
+		}
 
 		/* Bind it... */
 		if (bind(nsd->tcp[i].s, (struct sockaddr *) nsd->tcp[i].addr->ai_addr, nsd->tcp[i].addr->ai_addrlen) != 0) {
@@ -478,7 +490,7 @@ server_init(struct nsd *nsd)
 		log_msg(LOG_ERR, "The diff file contains errors. Will continue without it");
 	}
 #ifdef NSEC3
-	prehash(nsd->db, NULL);
+	prehash(nsd->db, 0);
 #endif
 
 	compression_table_capacity = 0;
@@ -693,10 +705,7 @@ server_reload(struct nsd *nsd, region_type* server_region, netio_type* netio,
 		region_log_stats(nsd->db->region);
 #endif
 #ifdef NSEC3
-	for(zone= nsd->db->zones; zone; zone = zone->next) {
-		if(zone->updated)
-			prehash(nsd->db, zone);
-	}
+	prehash(nsd->db, 1);
 #endif
 
 	initialize_dname_compression_tables(nsd);
@@ -1412,13 +1421,13 @@ handle_tcp_reading(netio_type *netio,
 		 *   + Query type         (2)
 		 */
 		if (data->query->tcplen < QHEADERSZ + 1 + sizeof(uint16_t) + sizeof(uint16_t)) {
-			log_msg(LOG_WARNING, "dropping bogus tcp connection");
+			VERBOSITY(2, (LOG_WARNING, "packet too small, dropping tcp connection"));
 			cleanup_tcp_handler(netio, handler);
 			return;
 		}
 
 		if (data->query->tcplen > data->query->maxlen) {
-			log_msg(LOG_ERR, "insufficient tcp buffer, dropping connection");
+			VERBOSITY(2, (LOG_WARNING, "insufficient tcp buffer, dropping connection"));
 			cleanup_tcp_handler(netio, handler);
 			return;
 		}
@@ -1537,6 +1546,9 @@ handle_tcp_writing(netio_type *netio,
 				 */
 				return;
 			} else {
+#ifdef ECONNRESET
+				if(verbosity>=2 || errno != ECONNRESET)
+#endif
 				log_msg(LOG_ERR, "failed writing to tcp: %s", strerror(errno));
 				cleanup_tcp_handler(netio, handler);
 				return;
@@ -1568,6 +1580,9 @@ handle_tcp_writing(netio_type *netio,
 			 */
 			return;
 		} else {
+#ifdef ECONNRESET
+			if(verbosity>=2 || errno != ECONNRESET)
+#endif
 			log_msg(LOG_ERR, "failed writing to tcp: %s", strerror(errno));
 			cleanup_tcp_handler(netio, handler);
 			return;
@@ -1660,7 +1675,15 @@ handle_tcp_accept(netio_type *netio,
 	addrlen = sizeof(addr);
 	s = accept(handler->fd, (struct sockaddr *) &addr, &addrlen);
 	if (s == -1) {
-		if (errno != EINTR) {
+		/* EINTR is a signal interrupt. The others are various OS ways
+		   of saying that the client has closed the connection. */
+		if (	errno != EINTR 
+			&& errno != EWOULDBLOCK 
+			&& errno != ECONNABORTED 
+#ifdef EPROTO
+			&& errno != EPROTO
+#endif /* EPROTO */
+			) {
 			log_msg(LOG_ERR, "accept failed: %s", strerror(errno));
 		}
 		return;
